//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "UaMain.h"
#include "UaConf.h"
#include "ControlQueue.h"
#include "Command.h"
#include "CallbackQueue.h"
#include "Callback.h"
#include "Settings.h"
#include <re.h>
#include "baresip.h"
#include "Log.h"
#include "Utils.h"
#include <assert.h>

#pragma package(smart_init)

//---------------------------------------------------------------------------
class Worker : public TThread
{
private:
protected:
	void __fastcall Execute();
public:
	__fastcall Worker(bool CreateSuspended);
};


#define DEBUG_MODULE "UaMain"
#define DEBUG_LEVEL 6
#include <re_dbg.h>

static struct {
	uint32_t n_uas;       /**< Number of User Agents           */
	bool terminating;     /**< Application is terminating flag */
	struct ua *ua;
	struct call *callp;	
} app;

struct ua* ua_cur(void) {
	return app.ua;
}

static const char software[] = "tSIP 0.1";

static bool appRestart = false;
static bool appQuit = false;

static void ua_exit_handler(int err)
{
	(void)err;
	re_cancel();
}

static void ua_event_handler(struct ua *ua, enum ua_event ev,
	struct call *call, const char *prm, void *arg)
{
	Callback::ua_state_e state;
	Callback::reg_state_e reg_state;
	/** \todo acc_id */
	switch (ev)
	{
	case UA_EVENT_CALL_INCOMING:
		state = Callback::CALL_STATE_INCOMING;
		app.callp = call;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_CALL_RINGING:
		state = Callback::CALL_STATE_RINGING;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_CALL_PROGRESS:
		state = Callback::CALL_STATE_PROGRESS;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_CALL_ESTABLISHED:
		state = Callback::CALL_STATE_ESTABLISHED;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_CALL_CLOSED:
		state = Callback::CALL_STATE_CLOSED;
		UA_CB->ChangeCallState(state, prm);
		app.callp = NULL;
		break;
	case UA_EVENT_REGISTERING:
		reg_state = Callback::REG_STATE_REGISTERING;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_REGISTER_OK:
		reg_state = Callback::REG_STATE_REGISTER_OK;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_REGISTER_FAIL:
		reg_state = Callback::REG_STATE_REGISTER_FAIL;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_UNREGISTERING:
		reg_state = Callback::REG_STATE_UNREGISTERING;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_UNREGISTER_OK:
		reg_state = Callback::REG_STATE_UNREGISTER_OK;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_UNREGISTER_FAIL:
		reg_state = Callback::REG_STATE_UNREGISTER_FAIL;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	default:
		assert(!"Unhandled UA event");
	}

}

static int ua_add(const struct pl *addr)
{
	char buf[512];
	int err;

	pl_strcpy(addr, buf, sizeof(buf));
	/** \note Do not pass stack variable as a ua argument, its value
		would be overwritten on object destruction! */
	err = ua_alloc(&app.ua, buf);
	if (err)
		return err;

	return err;
}

extern "C" void on_log(int level, const char *p)
{
	LOG("%s", p);
}

extern "C" int dialog_info_handler(int id, enum dialog_info_status status)
{
    UA_CB->ChangeDlgInfoState(id, status);
	return 0;
}

static int app_init(void)
{
	uint32_t n;
	int err;

    memset(&app, 0, sizeof(app));

	print_handler_set(on_log);

	/* Initialise System library */
	err = libre_init();
	if (err)
		return err;

	dbg_init(DBG_INFO, DBG_NONE);

	/* Initialise dynamic modules */
	mod_init();
#if 0
	static AnsiString asConfPath = ExtractFileDir(Application->ExeName).c_str();
	conf_path_set(asConfPath.c_str());
#endif
	struct config * cfg = conf_config();

    cfg->sound_backend = (config::e_sound_backend)appSettings.uaConf.soundBackend;

	strncpy(cfg->sip.local, appSettings.uaConf.local.c_str(), sizeof(cfg->sip.local));
	cfg->sip.local[sizeof(cfg->sip.local)-1] = '\0';

	cfg->codecs.g711 = appSettings.uaConf.codecs.g711;
	cfg->codecs.g726_32 = appSettings.uaConf.codecs.g726_32;
	cfg->codecs.gsm = appSettings.uaConf.codecs.gsm;
	cfg->codecs.speex = appSettings.uaConf.codecs.speex;

    configure();

	n = list_count(aucodec_list());
	LOG("Populated %u audio codec%s\n", n, 1==n?"":"s");
	if (0 == n) {
		DEBUG_WARNING("No audio-codec modules loaded!\n");
	}
	n = list_count(aufilt_list());
	LOG("Populated %u audio filter%s\n", n, 1==n?"":"s");

	/* Initialise User Agents */
	AnsiString uaName;
	uaName.sprintf("tSIP %s", GetFileVer(Application->ExeName).c_str());
	err = ua_init(uaName.c_str(), appSettings.uaConf.logMessages,
		true, true, true, false);
	if (err)
		return err;

	uag_event_register(ua_event_handler, NULL);

	return 0;
}


static int app_start(void)
{
	uint32_t n;

	app.n_uas = 0;
	for (int i=0; i<appSettings.uaConf.accounts.size(); i++)
	{
		AnsiString addr;
		UaConf::Account &acc = appSettings.uaConf.accounts[i];
		addr.sprintf("<sip:%s:%s@%s;transport=%s>;regint=%d", //;sipnat=outbound",
			acc.user.c_str(), acc.pwd.c_str(), acc.reg_server.c_str(),
			acc.getTransportStr(),
			acc.reg_expires
			);
		if (acc.auth_user != "")
		{
			addr.cat_printf(";auth_user=%s", acc.auth_user.c_str());
		}

		if (acc.stun_server != "")
		{
			addr.cat_printf(";medianat=stun;stunserver=stun:%s",
				acc.stun_server.c_str());
		}

		pl pl_addr;
		pl_set_str(&pl_addr, addr.c_str());
		if (ua_add(&pl_addr) == 0)
		{
			app.n_uas++;
		}
	}
	if (!app.n_uas)
	{
		DEBUG_WARNING("No SIP accounts found - check your config\n");
		return ENOENT;
	}

	if (appSettings.uaConf.accounts.size() != 1)
	{
		DEBUG_WARNING("Multiple accounts not handled!");
	}
	if (appSettings.uaConf.accounts.size() > 0)
	{
		UaConf::Account &acc = appSettings.uaConf.accounts[0];
		for (int i=0; i<appSettings.uaConf.contacts.size(); i++) {
			UaConf::Contact contact = appSettings.uaConf.contacts[i];
			if (contact.user == "")
			{
				continue;
			}
			AnsiString addr;
			if (contact.user.find("sip:") != std::string::npos)
			{
				addr.sprintf("<%s>", contact.user.c_str());
			}
			else
			{
				addr.sprintf("<sip:%s@%s;transport=%s>",
					contact.user.c_str(),
					acc.reg_server.c_str(),
					acc.getTransportStr()
					);
			}
			if (contact.sub_dialog_info)
			{
				addr.cat_printf(";dlginfo=p2p");
			}
			pl pl_addr;
			pl_set_str(&pl_addr, addr.c_str());
			contact_add(NULL, &pl_addr, i, dialog_info_handler);
		}
	}

	// contact list must be initialized here
	struct pl modname;
	pl_set_str(&modname, "dialog-info");
	load_module2(NULL, &modname);

	return 0;
}


static void app_close(void)
{
	ua_close();
	mod_close();
	libre_close();

	/* Check for memory leaks */
    LOG("Checking for memory leaks...");
	tmr_debug();
	mem_debug();
}

extern "C" void dialog_info_subscriber_close(void);

static void quit(int sig)
{
	if (!app.terminating) {
		/** \todo Closing dialog-info subscriber here shouldn't be necessary? */
		dialog_info_subscriber_close();	
		ua_stop_all(false);
		app.terminating = true;
		return;
	}

	LOG("forced exit - signal %d\n", sig);

	ua_stop_all(true);
}


extern "C" void signal_handler(int sig)
{
	LOG("terminated by signal %d\n", sig);
	quit(sig);
}


extern "C" int stderr_handler(const char *p, size_t size, void *arg)
{
	(void)arg;

	if (1 != fwrite(p, size, 1, stderr))
		return ENOMEM;

	return 0;
}

extern "C" void control_handler(void)
{
	if (app.terminating)
	{
		return;
	}
	else if (appQuit || appRestart)
	{
		//if (appQuit)
		//{
		//	app.terminating = true;
		//}
		quit(0);
	}
	
	int err;
	Command cmd;
	if (UA->GetCommand(cmd))
	{
    	return;
	}
	switch (cmd.type)
	{
	case Command::CALL:
		err = ua_connect(ua_cur(), &app.callp, NULL /*from*/,
			cmd.target.c_str(), NULL, VIDMODE_OFF);
		if (err)
		{
			DEBUG_WARNING("connect failed: %s\n", strerror(err));
		}
		break;
	case Command::ANSWER:
		ua_answer(ua_cur(), app.callp);
		break;
	case Command::TRANSFER:
		if (app.callp)
		{
			call_transfer(app.callp, cmd.target.c_str());
		}
		break;
	case Command::SEND_DIGIT:
		if (app.callp)
		{
			if (call_send_digit(app.callp, cmd.key) == 0)
			{
				call_send_digit(app.callp, 0x00);
			}
		}
		break;
	case Command::HOLD:
		if (app.callp)
		{
			call_hold(app.callp, cmd.bEnabled);
		}
		break;
	case Command::HANGUP:
		if (app.callp)
		{
			ua_hangup(ua_cur(), app.callp, cmd.code, NULL);
			app.callp = NULL;
		}
		break;
	case Command::SET_MSG_LOGGING:
		ua_log_messages(cmd.bEnabled);
		break;
	case Command::REREGISTER: {
		struct ua* ua = ua_cur();
		err |= ua_reregister(ua);
		break;
	}
	default:
		assert(!"Unhandled command type");
		break;
	}
}

static bool app_terminated = false;

__fastcall Worker::Worker(bool CreateSuspended)
	: TThread(CreateSuspended)
{
}
//---------------------------------------------------------------------------
void __fastcall Worker::Execute()
{
	int err;

	do
	{
		appRestart = false;
		err = app_init();
		if (err) {
			DEBUG_WARNING("app_init failed (%s)\n", strerror(err));
			goto out;
		}

		err = app_start();
		if (err) {
			DEBUG_WARNING("app_start failed (%s)\n", strerror(err));
			UA_CB->ChangeAppState(Callback::APP_START_FAILED);
			goto out;
		}

		/* Main loop */
		//while (!app.terminating) {
			err = re_main(signal_handler, control_handler);
		//}

	 out:
		app_close();
		list_flush(contact_list());		

		if (err) {
			DEBUG_WARNING("main exit with error (%s)\n", strerror(err));
			ua_stop_all(true);			
		} else {
			LOG("main exit w/o error, waiting for restart or termination\n");
		}
		while (!appRestart && !appQuit && !Terminated)
		{
			Sleep(10);
		}
	} while (appRestart);
	//return err;
	app_terminated = true;
}
//---------------------------------------------------------------------------

int Ua::Start(void)
{
	thread = new Worker(false);
	thread->FreeOnTerminate = true;
	return 0;
}

void Ua::Restart(void)
{
	appRestart = true;
}

void Ua::Quit(void)
{
	if (appQuit)
	{
    	return;
	}
	appQuit = true;
	while (!app_terminated)
	{
		Sleep(10);
		Application->ProcessMessages();
	}
}



